import atexit
import datetime as dt
import json
import logging
import logging.config
from pathlib import Path
from typing import override

PROJECT_ROOT = Path(__file__).parent
LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


def setup_logging(
        logdir: Path | str | None = None,
        **kwargs,
):
    """
    Setup logging configuration
    :param logdir: Directory to store the log file
    :keyword file_log_level: Log level for the file handler
    :keyword stdout_log_level: Log level for the stdout handler
    :return:
    """
    handlers = ["file", "stdout"]

    config_file = PROJECT_ROOT / "resources" / "logging_config.json"
    with open(config_file, "r") as file:
        config = json.load(file)

    # Override log level if provided
    for handler in handlers:
        if log_level := kwargs.get(f"{handler}_log_level"):
            level_str = logging.getLevelName(log_level.upper())
            if not isinstance(level_str, str):
                level_str = logging.getLevelName(level_str)
            config["handlers"][handler]["level"] = level_str

    # Set log file path to user log directory
    if logdir:
        logdir = Path(logdir)
        config["handlers"]["file"]["filename"] = logdir / "adGroupSync.log.jsonl"

    # Create path and file if it does not exist
    Path(config["handlers"]["file"]["filename"]).parent.mkdir(
        parents=True, exist_ok=True)
    Path(config["handlers"]["file"]["filename"]).touch()

    logging.config.dictConfig(config)
    queue_handler = logging.getHandlerByName("queue_handler")
    if queue_handler is not None:
        queue_handler.listener.start()
        atexit.register(queue_handler.listener.stop)


class JSONFormatter(logging.Formatter):
    """
    A custom JSON formatter for logging
    """
    HIDE_KEYS = ["password", "token", "api_key", "site_key"]


    def __init__(
            self,
            *,
            fmt_keys: dict[str, str] | None = None,
    ):
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else {}

    @override
    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)

        # Exclude passwords from the log
        self._hide_passwords(record)

        return json.dumps(message, default=str)

    def _hide_passwords(self, log_record: logging.LogRecord|dict):
        """
        Recursively replace all values with keys containing 'password',
        'token', etc. with '********'
        :param log_record:
        :return:
        """
        if not isinstance(log_record, dict):
            dict_obj = log_record.__dict__
        else:
            dict_obj = log_record

        for key, value in dict_obj.items():
            if isinstance(value, dict):
                dict_obj = self._hide_passwords(value)
            elif any(hide_key in key.lower() for hide_key in self.HIDE_KEYS):
                dict_obj[key] = "********"

        if isinstance(log_record, logging.LogRecord):
            for key, value in dict_obj.items():
                setattr(log_record, key, value)
            return log_record
        else:
            return dict_obj

    def _prepare_log_dict(self, record: logging.LogRecord) -> dict:
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created, tz=dt.timezone.utc
            ).isoformat()
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: msg_val
            if (msg_val := always_fields.pop(val, None)) is not None
            else getattr(record, val)
            for key, val in self.fmt_keys.items()
        }
        message.update(always_fields)

        # Include all other attributes
        for key, val, in record.__dict__.items():
            if key not in LOG_RECORD_BUILTIN_ATTRS:
                message[key] = val

        return message